/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is a test framework for NEON YUV-to-ARGB code.
 *
 * The Initial Developer of the Original Code is ARM Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jacob Bramley <Jacob.Bramley@arm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>

#include "framework.h"

#include "runner.h"

char const * me = NULL;     // The name of this program (taken from argv[0]).

void usage(void)
{
    printf("\n");
    printf("Usage: %s [options] <RAW YCbCr file>\n", me);
    printf("\n");
    printf("If the input file has the following format, configuration options will be\n");
    printf("determined automatically:  <base_name>-<width>x<height>[-pad<pad>].yuv\n");
    printf("Alternatively, they may be specified manually using the command-line options.\n");
    printf("Command-line options always override options derived from the file name.\n");
    printf("\n");
    printf("Note that the option parser is primitive and inflexible. The arguments must be\n");
    printf("specified with an \"=\" symbol between the option and the argument, everything is\n");
    printf("case-sensitive, and there are no short-form alternatives for any of the arguments.\n");
    printf("\n");
    printf("Options:\n");
    printf("  --format=<format>\n");
    printf("Supported formats are:\n");
    printf("     IYUV (default)\n");
    printf("     I420 (equivalent to IYUV)\n");
    printf("     YV12\n");
    printf("  --width=<width>\n");
    printf("  --height=<height>\n");
    printf("  --scale=<scale>\n");
    printf("  --scale-x=<scale>\n");
    printf("  --scale-y=<scale>\n");
    printf("        The scale value can be a floating-point argument. The --scale option sets both\n");
    printf("        sub-options at once.\n");
    printf("  --out-width=<width>\n");
    printf("  --out-height=<height>\n");
    printf("        The --out-width and --out-height arguments define the minimum size of the output.\n");
    printf("        These arguments can be used to test that the conversion routines do not over-write.\n");
    printf("        The unused area will be filled with a standard transparency grid pattern. The result\n");
    printf("        will be centred on the output canvas.\n");
    printf("  --out-left=<left>\n");
    printf("  --out-top=<top>\n");
    printf("        These arguments can be used to locate the output image in the output canvas.\n");
    printf("  --pad=<pad>\n");
    printf("  --pad-top=<pad>\n");
    printf("  --pad-right=<pad>\n");
    printf("  --pad-bottom=<pad>\n");
    printf("  --pad-left=<pad>\n");
    printf("        The padding options instruct the converter to ignore <pad> pixels around the \n");
    printf("        edge of the source. The --pad option sets all four sub-options at once.\n");
    printf("        Note that the <width> and <height> options do not include the padding size, so\n");
    printf("        the source must have (width+pad_left+pad_right)*(height+pad_top+pad_bottom)\n");
    printf("        pixels.\n");
    printf("  --runs=<n>\n");
    printf("        Run the conversion routines <n> times and report the average time taken.\n");
    printf("        If this is set to 0, which is the default, the conversion will be run once, but no\n");
    printf("        timing information will be collected.\n");
    printf("  --verbose\n");
    printf("        Turn on verbose mode.\n");
    printf("\n");
}

#define OPT_FORMAT          "--format"
#define OPT_WIDTH           "--width"
#define OPT_HEIGHT          "--height"
#define OPT_SCALE           "--scale"
#define OPT_SCALE_X         "--scale-x"
#define OPT_SCALE_Y         "--scale-y"
#define OPT_OUT_WIDTH       "--out-width"
#define OPT_OUT_HEIGHT      "--out-height"
#define OPT_OUT_LEFT        "--out-left"
#define OPT_OUT_TOP         "--out-top"
#define OPT_PAD             "--pad"
#define OPT_PAD_TOP         "--pad-top"
#define OPT_PAD_RIGHT       "--pad-right"
#define OPT_PAD_BOTTOM      "--pad-bottom"
#define OPT_PAD_LEFT        "--pad-left"
#define OPT_RUNS            "--runs"
#define OPT_VERBOSE         "--verbose"
#define OPT_HELP            "--help"

// opt_: One of the OPT_ strings.
// arg_: A pointer to the argument, as received on the command line.
// value_: Somewhere to store the parsed value.
// func_: The function to use to parse the value.
#define PARSE_OPT(opt_, arg_, value_, func_)                        \
    if (strstr((arg_), (opt_)) == (arg_)) {                         \
        if (!func_((opt_), (value_), (arg_)+strlen(opt_))) {        \
            goto exit;                                              \
        }                                                           \
        continue;                                                   \
    }

// Option-parsing functions. These all update the options structure based on
// the string input. Parse failures (etc) will be indiciated by returning 0.
// Successful calls will return 1.
static int parse_format_quiet(char const * opt, format_t * value, char const * arg, int quiet)
{
    if ((arg[0] != '=') || (arg[1] == '\0')) {
        printf("Error: \"%s\" requires a value.\n", opt);
        return 0;
    }
    arg++;  // Skip past the '='.
    if (strcmp(arg, "IYUV") == 0) {
        *value = FMT_IYUV;
    } else if (strcmp(arg, "I420") == 0) {
        *value = FMT_I420;
    } else if (strcmp(arg, "YV12") == 0) {
        *value = FMT_YV12;
    } else {
        if (!quiet) {
            printf("Error: \"%s\" is not a recognized value for \"%s\".\n", arg, opt);
        }
        return 0;
    }
    return 1;
}

static int parse_format(char const * opt, format_t * value, char const * arg)
{
    return parse_format_quiet(opt, value, arg, 0);
}

static int parse_uint(char const * opt, uint32_t * value, char const * arg)
{
    int         n;
    uint32_t    v;
    if ((arg[0] != '=') || (arg[1] == '\0')) {
        printf("Error: \"%s\" requires a value.\n", opt);
        return 0;
    }
    arg++;  // Skip past the '='.
    if ((sscanf(arg, "%u%n", &v, &n) != 1) || (arg[n] != '\0')) {
        printf("Error: Expected an integer value for option \"%s\".\n", opt);
        return 0;
    }
    *value = v;
    return 1;
}

static int parse_float(char const * opt, float * value, char const * arg)
{
    int         n;
    float       v;
    if ((arg[0] != '=') || (arg[1] == '\0')) {
        printf("Error: \"%s\" requires a value.\n", opt);
        return 0;
    }
    arg++;  // Skip past the '='.
    if ((sscanf(arg, "%f%n", &v, &n) != 1) || (arg[n] != '\0')) {
        printf("Error: Expected a value for option \"%s\".\n", opt);
        return 0;
    }
    *value = v;
    return 1;
}

static int parse_bool(char const * opt, int * value, char const * arg)
{
    if (arg[0] != '\0') {
        printf("Error: \"%s\" is a boolean option and takes no argument.\n", opt);
        return 0;
    }
    *value = 1;
    return 1;
}

static int parse_pad(char const * opt, options_t * options, char const * arg)
{
    uint32_t    pad;
    int ret = parse_uint(opt, &pad, arg);
    options->pad_top = pad;
    options->pad_right = pad;
    options->pad_bottom = pad;
    options->pad_left = pad;
    return ret;
}

static int parse_scale(char const * opt, options_t * options, char const * arg)
{
    float   scale;
    int ret = parse_float(opt, &scale, arg);
    options->scale_x = scale;
    options->scale_y = scale;
    return ret;
}

// Parse the source file name, to try to populate some default options.
// This works for file names like this:
//  <base_name>-<width>x<height>[-pad<pad>].yuv
//
// This function returns '1' on success, or '0' if an error occurred. A failure
// to match does not constitute an error. However, if no match occurred, the
// options structure is not updated.
static int try_filename_parse(options_t * options, char const * name)
{
    int         ret = 0;
    uint32_t    width = options->src_width;
    uint32_t    height = options->src_height;
    uint32_t    pad = options->pad_top;
    uint32_t    name_len = strlen(name);
    uint32_t    base_name_len = name_len;
    int32_t     caret;
    uint32_t    field_count;
    char        fourcc[6] = "=";
    format_t    format;
    char *      base_name = malloc(name_len+1);
    OOM_CHECK(base_name);

    // Copy the string to the base_name buffer.
    strcpy(base_name, name);

    // Scan from the source name for image information.
    field_count = 0;
    for (caret = name_len-4; caret > 0; caret--) {
        int n;
        if (base_name[caret] != '-') {
            continue;
        }
        field_count++;
        if (field_count == 1) {
            // Try this: "<base_name>-<width>x<height>.<format>"
            if (sscanf(&base_name[caret], "-%ux%u.%4s%n", &width, &height, fourcc+1, &n) != 3) {
                // No match. Continue searching for the "-pad" form.
                continue;
            }
            if (caret + n != base_name_len) {
                // There was an incomplete match, so treat it as inconclusive.
                continue;
            }
            // Convert each fourcc character to upper case, for
            // case-insensitive matching.
            for (uint32_t i = 0; fourcc[i] != '\0'; i++) {
                if ((fourcc[i] >= 'a') && (fourcc[i] <= 'z')) {
                    fourcc[i] += 'A'-'a';
                }
            }
            // Try to match the fourcc to a known format.
            if (!parse_format_quiet("<format>", &format, fourcc, 1)) {
                continue;
            }
            // We've found a match.
            base_name_len = caret;
            base_name[caret] = '\0';
            pad = 0;
            goto set_options;
        }
        if (field_count == 2) {
            // Try this: "<base_name>-<width>x<height>-pad<pad>.<format>"
            if (sscanf(&base_name[caret], "-%ux%u-pad%u.%4s%n", &width, &height, &pad, fourcc+1, &n) != 4) {
                // No match.
                goto exit_ok;
            }
            if (caret + n != base_name_len) {
                // There was an incomplete match, so treat it as inconclusive.
                goto exit_ok;
            }
            // Convert each fourcc character to upper case, for
            // case-insensitive matching.
            for (uint32_t i = 0; fourcc[i] != '\0'; i++) {
                if ((fourcc[i] >= 'a') && (fourcc[i] <= 'z')) {
                    fourcc[i] += 'A'-'a';
                }
            }
            // Try to match the fourcc to a known format.
            if (!parse_format_quiet("<format>", &format, fourcc, 1)) {
                continue;
            }
            // We've found a match.
            base_name_len = caret;
            base_name[caret] = '\0';
            goto set_options;
        }
        // Too many fields, so no match.
        goto exit_ok;
    }

set_options:            // Options for known values.
    options->src_width = width;
    options->src_height = height;
    options->pad_top = pad;
    options->pad_right = pad;
    options->pad_bottom = pad;
    options->pad_left = pad;
    options->format = format;
    options->base_name = malloc(base_name_len+1);
    options->base_name_len = base_name_len;
    strcpy(options->base_name, base_name);

exit_ok:
    ret = 1;
exit:
    free(base_name);
    return  ret;
}

int main(int argc, char const * argv[])
{
    uint32_t        i;
    int             ret = EXIT_FAILURE;
    char const **   args = NULL;
    char const **   arg = NULL;
    options_t       options;

    // Set default options.
    options.base_name = NULL;
    options.base_name_len = 0;
    options.src_file = NULL;    // This must always be specified.
    options.format = FMT_IYUV;
    options.src_width = 0;      // This must always be specified.
    options.src_height = 0;     // This must always be specified.
    options.pad_top = 0;
    options.pad_right = 0;
    options.pad_bottom = 0;
    options.pad_left = 0;
    options.dst_width = 0;      // No minimum size by default.
    options.dst_height = 0;     // No minimum size by default.
    options.dst_x = 0;
    options.dst_y = 0;
    options.scale_x = 1.0f;
    options.scale_y = 1.0f;
    options.runs = 0;           // '0' denotes a testing run.
    options.verbose = 0;

    me = argv[0];
    if (argc < 2) {
        usage();
        goto exit;
    }

    // Exactly one option will be a file name, and exactly one is the
    // executable name. Any others will be real options. We also append a NULL
    // pointer to act as a terminator.
    args = malloc(sizeof(char const *) * (argc-1));
    OOM_CHECK(args);

    arg = args;
    for (uint32_t i = 1; i < argc; i++) {
        // Check for arguments.
        if (strstr(argv[i], "--") == argv[i]) {
            *arg = argv[i];
            arg++;
            continue;
        }
        // Check for the source file.
        if (options.src_file) {
            printf("\nError: Source file \"%s\" specified, but \"%s\" was already given.\n",
                    argv[i],
                    options.src_file);
            usage();
            goto exit;
        }
        options.src_file = argv[i];
    }
    *arg = NULL; // Terminate the args list.

    if (!options.src_file) {
        printf("\nError: No source file specified.\n");
        usage();
        goto exit;
    }

    // Parse the source file name, to try to populate some default options.
    // This works for file names like this:
    //  <base_name>-<width>x<height>[-pad<pad>].yuv
    if (!try_filename_parse(&options, options.src_file)) {
        goto exit;
    }

    // Parse remaining command-line arguments.
    for (arg = args; *arg != NULL; arg++) {
        PARSE_OPT(OPT_FORMAT,       *arg,   &options.format,        parse_format);
        PARSE_OPT(OPT_WIDTH,        *arg,   &options.src_width,     parse_uint);
        PARSE_OPT(OPT_HEIGHT,       *arg,   &options.src_height,    parse_uint);
        PARSE_OPT(OPT_SCALE_X,      *arg,   &options.scale_x,       parse_float);           
        PARSE_OPT(OPT_SCALE_Y,      *arg,   &options.scale_y,       parse_float);           
        PARSE_OPT(OPT_SCALE,        *arg,   &options,               parse_scale);
        PARSE_OPT(OPT_OUT_WIDTH,    *arg,   &options.dst_width,     parse_uint);
        PARSE_OPT(OPT_OUT_HEIGHT,   *arg,   &options.dst_height,    parse_uint);
        PARSE_OPT(OPT_OUT_LEFT,     *arg,   &options.dst_x,         parse_uint);
        PARSE_OPT(OPT_OUT_TOP,      *arg,   &options.dst_y,         parse_uint);
        PARSE_OPT(OPT_PAD_TOP,      *arg,   &options.pad_top,       parse_uint);
        PARSE_OPT(OPT_PAD_RIGHT,    *arg,   &options.pad_right,     parse_uint);
        PARSE_OPT(OPT_PAD_BOTTOM,   *arg,   &options.pad_bottom,    parse_uint);
        PARSE_OPT(OPT_PAD_LEFT,     *arg,   &options.pad_left,      parse_uint);
        PARSE_OPT(OPT_PAD,          *arg,   &options,               parse_pad);
        PARSE_OPT(OPT_RUNS,         *arg,   &options.runs,          parse_uint);
        PARSE_OPT(OPT_VERBOSE,      *arg,   &options.verbose,       parse_bool);

        if (strstr(*arg, OPT_HELP) == *arg) {
            usage();
            ret = EXIT_SUCCESS;
            goto exit;
        }

        printf("Error: Unrecognized argument: \"%s\"\n", *arg);
        goto exit;
    }

    // Ensure that the output canvas is big enough to hold the scaled image.
    i = options.dst_x + (uint32_t)(options.scale_x * options.src_width + 0.5f);
    if (i > options.dst_width) {
        options.dst_width = i;
    }
    i = options.dst_y + (uint32_t)(options.scale_y * options.src_height + 0.5f);
    if (i > options.dst_height) {
        options.dst_height = i;
    }

    if (options.verbose) {
        printf("  Verbose mode is ON.\n");
        printf("  Source file: %s\n", options.src_file);
        printf("  Source size: %ux%u pixels\n", options.src_width, options.src_height);
        printf("  Source padding: %u (top) %u (right) %u (bottom) %u (left)\n",
                options.pad_top,
                options.pad_right,
                options.pad_bottom,
                options.pad_left);
        printf("  Scale factor: %f horizontally, %f vertically.\n",
                options.scale_x,
                options.scale_y);
        if (options.runs) {
            printf("  Benchmark runs: %u\n", options.runs);
        }
        printf("  Output canvas size: %ux%u pixels\n", options.dst_width, options.dst_height);
        printf("  Output base name: %s\n", options.base_name);
    }

    if (!run(&options)) {
        goto exit;
    }

    ret = EXIT_SUCCESS;
exit:
    free(args);
    free(options.base_name);
    return ret;
}

